/*
 * Copyright (c) 2024. Helge Eichhorn and the LOX contributors
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, you can obtain one at https://mozilla.org/MPL/2.0/.
 */

//! Mod `units` exposes shared type aliases representing a variety of physical units.

pub type Arcseconds = f64;

pub type Microarcseconds = f64;

pub type Radians = f64;

pub type Seconds = f64;

pub type JulianCenturies = f64;

pub type Days = f64;
